/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "core.h"
#include "resources.h"

namespace mysdl
{
	namespace audio
	{
		class AudioResourceBankGenerator : public ResourceBankGenerator
		{
			public:
				BasicResourceBank *generateBank(const std::string &name) const;
		};

		BasicResourceBank *AudioResourceBankGenerator::generateBank(const std::string &name) const
		{
			if(name=="samples")
			{
				return(new audio::SampleBank());
			}
			else
			{
				return(0);
			}
		}

		AudioResourceBankGenerator resourceBankGenerator;

		bool enabled=true;
		ALCdevice *device=0;
		Context *context=0;

		std::string getALErrorString(ALenum error)
		{
			switch(error) // \todo complete
			{
				case AL_NO_ERROR:
					return(std::string("AL_NO_ERROR"));

				case AL_INVALID_NAME:
					return(std::string("AL_INVALID_NAME"));

				case AL_INVALID_VALUE:
					return(std::string("AL_INVALID_VALUE"));

				case AL_OUT_OF_MEMORY:
					return(std::string("AL_OUT_OF_MEMORY"));

				default:
					return(std::string("(Unknown AL error)"));
			};
		}

		std::string getALCErrorString(ALenum error)
		{
			switch(error) // \todo complete
			{
				case ALC_NO_ERROR:
					return(std::string("AL_NO_ERROR"));

				case ALC_INVALID_DEVICE:
					return(std::string("ALC_INVALID_DEVICE"));

				case ALC_INVALID_CONTEXT:
					return(std::string("ALC_INVALID_CONTEXT"));

				case ALC_INVALID_ENUM:
					return(std::string("ALC_INVALID_ENUM"));

				case ALC_INVALID_VALUE:
					return(std::string("ALC_INVALID_VALUE"));

				case ALC_OUT_OF_MEMORY:
					return(std::string("ALC_OUT_OF_MEMORY"));

				default:
					return(std::string("(Unknown ALC error)"));
			};
		}

		void init(std::string deviceName,int frequency)
		{
			context=0;

/*			if(!resourceBankGenerator)
			{
				resourceBankGenerator=new AudioResourceBankGenerator();
				registerResourceBankGenerator(resourceBankGenerator);
			}*/

			if(enabled)
			{
				alcGetError(0);
				device=alcOpenDevice(deviceName.length() ? (ALCchar *)(deviceName.c_str()) : 0);//device=alcOpenDevice(deviceName.length() ? (ALubyte *)(deviceName.c_str()) : 0);

				if(!device)
				{
					throw getALCErrorString(alcGetError(0));
				}

 				debug::debug(dbgAudio,"Initialized OpenAL device \"%s\".",alcGetString(device,ALC_DEVICE_SPECIFIER));
			}

			new Context(frequency);
		}

		void cleanUp()
		{
			if(context)
			{
				delete context;
				context=0;
			}

			if(device)
			{
				alcCloseDevice(device);
				device=0;
			}

			/*if(resourceBankGenerator)
			{
				unregisterResourceBankGenerator(resourceBankGenerator);
				delete resourceBankGenerator;
				resourceBankGenerator=0;
			}*/
		}

		Context::Context(int frequency)
		{
			if(context)
			{
				throw std::string("Cannot create more than one audio context.");
			}

			int attributes[3]={ALC_FREQUENCY,frequency,ALC_INVALID};

			//debug::addMessage("Context::Context()");

			alcContext=0;

			if(enabled)
			{
				alcGetError(device);
				alcContext=alcCreateContext(device,attributes);

				if(!alcContext)
				{
					//debug::removeMessage();
					throw getALCErrorString(alcGetError(device));
				}

				alcGetError(device);

				alcMakeContextCurrent(alcContext);

				ALenum error=alcGetError(device);

				if(error!=AL_NO_ERROR)
				{
					debug::warning(dbgAudio,getALCErrorString(error).c_str());
				}
			}

			context=this;

//			defaultListener=new Listener();
//			defaultListener->apply();

			ALfloat pos[]={0.0,0.0,0.0};
			ALfloat orientation[]={0.0,0.0,-1.0,0.0,1.0,0.0};

			alListenerfv(AL_POSITION,pos);
			alListenerfv(AL_ORIENTATION,orientation);

			//debug::removeMessage();
		}

		Context::~Context()
		{
			//debug::addMessage("Context::~Context()");

// 			for(std::list<Emitter *>::iterator itr=emitters.begin(); itr!=emitters.end(); itr++)
// 			{
// 				debug::debug(dbgAudio,"Stopping emitter %p.",(*itr));
// 				(*itr)->stop();
// 			}

//			perform(0);

// 			debug::debug(dbgAudio,"Context clean, destroying context %p.",this);

// 			if(defaultListener)
// 			{
// 				delete defaultListener;
// 				defaultListener=0;
// 			}

			if(alcContext)
			{
				alcDestroyContext(alcContext);
			}

			//debug::removeMessage();
		}

		void Context::perform(GLfloat ticks)
		{
			//debug::addMessage("Context::perform()");

// 			for(std::list<Emitter *>::iterator itr=emitters.begin(); itr!=emitters.end(); itr++)
// 			{
// 				(*itr)->perform(ticks);
// 			}

			//debug::removeMessage();
		}
	}
};
